from doipy.actions.fdo_config_type_14 import create_fdo_config_type_14
from doipy.constants import Profile
from doipy.exceptions import ProfileNotSupportedException
from doipy.models import FdoInput
from doipy.validation_utils import general_validation, get_specific_schema


def create_fdo(user_input: dict):

    # validate the user input against the general input schema
    general_validation(user_input)

    # get the schema which belongs to the inputs of the create operation for the chosen profile
    specific_schema = get_specific_schema(user_input)

    # validate the user input against the input schema for the specific profile and create an instance of FdoInput class
    fdo_input = FdoInput.parse(user_input, specific_schema)

    # get the profile from the FdoInput
    profile = fdo_input.fdo_profile_ref

    # choose the correct create_fdo operation
    if profile == Profile.CONFIG_TYPE_14.value:
        response = create_fdo_config_type_14(fdo_input)
    # check here for other profiles/configuration types
    else:
        raise ProfileNotSupportedException('Create_FDO is not supported by DOIPY for the chosen Profile.')

    # add more create_fdo operations for other configuration types
    return response



import json
from pathlib import Path
from typing import Annotated
import typer
from jsonschema.exceptions import ValidationError

from doipy import get_connection, hello, list_operations, create, retrieve, delete, search, create_fdo, get_design, \
    get_init_data
from doipy.constants import ResponseStatus, DOType
from doipy.exceptions import AuthenticationException, InvalidRequestException, ProfileNotSupportedException, \
    HandleValueNotFoundException, OperationNotSupportedException
from doipy.request_and_response import print_json_response

app = typer.Typer()


@app.command(name='hello')
def hello_command(target_id: Annotated[str, typer.Argument(help='the targetId identifying a data service')],
                  ip: Annotated[str, typer.Argument(help='the IP address identifying the data service')],
                  port: Annotated[int, typer.Argument(help='the DOIP port of the data service')]):
    """
    Implements 0.DOIP/Op.Hello: An operation to allow a client to get information about the DOIP service.
    """
    response = hello(target_id, ip, port)
    print_json_response(response)


@app.command(name='list_operations')
def list_operations_command(target_id: Annotated[str, typer.Argument(help='the targetId identifying a data service')],
                            ip: Annotated[str, typer.Argument(help='the IP address identifying a data service')],
                            port: Annotated[int, typer.Argument(help='the DOIP port of the data service')],
                            username: Annotated[
                                str, typer.Option(help='the username of the user at the data service')] = None,
                            client_id: Annotated[
                                str, typer.Option(help='the clientId of the user at the data service')] = None,
                            password: Annotated[
                                str, typer.Option(help='the password of the user at the data service')] = None,
                            token: Annotated[str, typer.Option(help='a token at the data service')] = None):
    """
    Implements 0.DOIP/Op.ListOperations: An operation to request the list of operations that can be invoked on the
    target DO.
    """
    try:
        response = list_operations(target_id, ip, port, username, client_id, password, token)
    except AuthenticationException as error:
        print(str(error))
        raise typer.Exit()

    print_json_response(response)


@app.command(name='create')
def create_command(target_id: Annotated[str, typer.Argument(help='the targetId identifying a data service')],
                   ip: Annotated[str, typer.Argument(help='the IP address identifying a data service')],
                   port: Annotated[int, typer.Argument(help='the DOIP port of the data service')],
                   do_type: Annotated[
                       str, typer.Option(help='name of the DO to be generated at the data service')] = DOType.DO.value,
                   do_name: Annotated[
                       str, typer.Option(help='name of the DO to be generated at data service')] = 'Digital Object',
                   bitsq: Annotated[
                       Path, typer.Option(
                           help='Path to a file which comprises the data bit-sequence of the DO to be generated')]
                   = None,
                   metadata: Annotated[
                       Path, typer.Option(
                           help='Path to a JSON file which comprises the data bit-sequence of the DO to be generated')]
                   = None,
                   username: Annotated[
                       str, typer.Option(help='the username of the user at the data service')] = None,
                   client_id: Annotated[
                       str, typer.Option(help='the clientId of the user at the data service')] = None,
                   password: Annotated[
                       str, typer.Option(help='the password of the user at the data service')] = None,
                   token: Annotated[str, typer.Option(help='a token at the data service')] = None):
    """
    Implements 0.DOIP/Op.Create: An operation to create a digital object (containing at most one data bit-sequence)
    within the DOIP service. The target of a creation operation is the DOIP service itself.
    """
    # get the metadata and write them into a dictionary
    metadata_dict = None
    if metadata:
        try:
            with open(metadata, 'r') as f:
                metadata_dict = json.load(f)
        except json.JSONDecodeError as error_message:
            print("Invalid JSON syntax:", error_message)
            raise typer.Exit()
        except Exception as error_message:
            print(error_message)
            raise typer.Exit()

    try:
        response = create(target_id, ip, port, do_type, do_name, bitsq, metadata_dict, username, client_id, password,
                          token)
    except AuthenticationException as error:
        print(str(error))
        raise typer.Exit()
    # in case that the DOIP response status is not success, raise an exception and terminate the program
    except InvalidRequestException as error:
        details = error.args[0][0]
        print(f'{details["status"]} {ResponseStatus(details["status"]).name}')
        print(details['output']['message'])
        raise typer.Exit()

    print_json_response(response)


@app.command(name='retrieve')
def retrieve_command(target_id: Annotated[str, typer.Argument(help='the targetId identifying an (F)do')],
                     ip: Annotated[str, typer.Argument(help='the IP address identifying a data service')],
                     port: Annotated[int, typer.Argument(help='the DOIP port of the data service')],
                     file: Annotated[str, typer.Option(help='If the bit-sequence should be returned, this is the id of '
                                                            'the file in the DOIP service')] = None,
                     username: Annotated[
                         str, typer.Option(help='the username of the user at the data service')] = None,
                     client_id: Annotated[
                         str, typer.Option(help='the clientId of the user at the data service')] = None,
                     password: Annotated[
                         str, typer.Option(help='the password of the user at the data service')] = None,
                     token: Annotated[str, typer.Option(help='a token at the data service')] = None):
    """
    Implements 0.DOIP/Op.Retrieve: An operation to allow a client to get information about an (F)DO at a service.
    """
    try:
        response = retrieve(target_id, ip, port, file, username, client_id, password, token)
    except AuthenticationException as error:
        print(str(error))
        raise typer.Exit()

    print_json_response(response)


@app.command(name='delete')
def delete_command(target_id: Annotated[str, typer.Argument(help='the targetId identifying an (F)do')],
                   ip: Annotated[str, typer.Argument(help='the IP address identifying a data service')],
                   port: Annotated[int, typer.Argument(help='the DOIP port of the data service')],
                   username: Annotated[
                       str, typer.Option(help='the username of the user at the data service')] = None,
                   client_id: Annotated[
                       str, typer.Option(help='the clientId of the user at the data service')] = None,
                   password: Annotated[
                       str, typer.Option(help='the password of the user at the data service')] = None,
                   token: Annotated[str, typer.Option(help='a token at the data service')] = None):
    """
    Implements 0.DOIP/Op.Delete: An operation to allow a client to delete an (F)DO at a service. This operation just
    deletes the referenced (F)DO but not any other (F)DOs which are referenced by the given (F)DO.
    """
    try:
        response = delete(target_id, ip, port, username, client_id, password, token)
    except AuthenticationException as error:
        print(str(error))
        raise typer.Exit()

    print_json_response(response)


@app.command(name='search')
def search_command(target_id: Annotated[str, typer.Argument(help='the targetId identifying an (F)do')],
                   ip: Annotated[str, typer.Argument(help='the IP address identifying a data service')],
                   port: Annotated[int, typer.Argument(help='the DOIP port of the data service')],
                   query: Annotated[
                       str, typer.Argument(help='The search query to be performed, in a textual representation')],
                   username: Annotated[
                       str, typer.Option(help='the username of the user at the data service')] = None,
                   client_id: Annotated[
                       str, typer.Option(help='the clientId of the user at the data service')] = None,
                   password: Annotated[
                       str, typer.Option(help='the password of the user at the data service')] = None,
                   token: Annotated[str, typer.Option(help='a token at the data service')] = None):
    """Implements 0.DOIP/Op.Search"""
    try:
        response = search(target_id, ip, port, query, username, client_id, password, token)
    except AuthenticationException as error:
        print(str(error))
        raise typer.Exit()

    print_json_response(response)


@app.command(name='create_fdo')
def create_fdo_command(
        input_file: Annotated[Path, typer.Argument(help='A file containing a JSON which follows a specific '
                                                        'JSON schema. The file contains data bit-sequences, '
                                                        'metadata bit-sequences and the metadata that should '
                                                        'be written into the corresponding PID records.')]):
    """
    Create a FAIR Digital Object (FDO).
    """
    # read the user input
    try:
        with open(input_file, 'r') as f:
            user_input = json.load(f)
    except json.JSONDecodeError as error_message:
        print("Invalid JSON syntax:", error_message)
        raise typer.Exit()
    except Exception as error_message:
        print(error_message)
        raise typer.Exit()

    # Create the FDO
    try:
        response = create_fdo(user_input)
    except ValidationError as error:
        print(str(error))
        raise typer.Exit()
    except OperationNotSupportedException as error:
        print(str(error))
        raise typer.Exit()
    # in case that authentication fails, raise an exception
    except AuthenticationException as error:
        print(str(error))
        raise typer.Exit()
    # in case the chosen profile is not supported, raise an exception
    except ProfileNotSupportedException as error:
        print(str(error))
        raise typer.Exit()
    # in case that the DOIP response is not success, raise an exception and terminate the program
    except InvalidRequestException as error:
        details = error.args[0][0]
        print(f'{details["status"]} {ResponseStatus(details["status"]).name}')
        print(details['output']['message'])
        raise typer.Exit()

    print_json_response(response)


@app.command(name='get_connection')
def get_connection_command(service: Annotated[str, typer.Argument(help='the PID identifying a data service')]):
    """
    Implements 0.DOIP/Op.Hello: An operation to allow a client to get information about the DOIP service.
    """
    try:
        service_id, ip, port = get_connection(service)
        print(f'Service ID: {service_id}')
        print(f'IP: {ip}')
        print(f'Port: {port}')
    except HandleValueNotFoundException as error:
        print(error)


@app.command(name='get_design')
def get_design_command():
    """Implements 20.DOIP/Op.GetDesign (see https://www.cordra.org)"""
    response = get_design()
    print_json_response(response)


@app.command(name='get_init_data')
def get_init_data_command():
    """Implements 20.DOIP/Op.GetInitData (see https://www.cordra.org)"""
    response = get_init_data()
    print_json_response(response)



import json
from doip_sdk import Response


def decode_json_response(response: Response) -> list:
    result = []
    for item in response.content:
        decoded_json = json.loads(item)
        result.append(decoded_json)
    return result


def print_json_response(response: list):
    for item in response:
        print(json.dumps(item, indent=2))
        print('#')
    print('#')



import requests

from doipy.constants import DomainName
from doipy.exceptions import HandleValueNotFoundException

cached_data = {}


def get_connection(service: str) -> (str, str, int):
    """
    Get the serviceId, IP and port from the handle correponding to the service that should be connected to.
    """
    # get required handle values from the handle record
    url = f'https://hdl.handle.net/api/handles/{service}'
    # cache the service_id, ip and port of the requested service
    if service not in cached_data:
        r = requests.get(url)
        data = r.json()
        # find ip and port in the handle values
        ip = ''
        port = 0
        ip_found = False
        port_found = False
        for item in data['values']:
            if item['type'] == 'IP':
                ip = item['data']['value']
                ip_found = True
            if item['type'] == 'PORT':
                port = item['data']['value']
                port_found = True
            if ip_found and port_found:
                break
        if not ip_found:
            raise HandleValueNotFoundException('Could not find IP in the Service Handle')
        elif not port_found:
            raise HandleValueNotFoundException('Could not find Port in the Service Handle')
        else:
            cached_data[service] = {
                'service_id': service,
                'ip': ip,
                'port': port
            }
    return cached_data[service]['service_id'], cached_data[service]['ip'], int(cached_data[service]['port'])


def get_service_id(fdo_service_ref: str) -> str:
    # get the service information from the dtr
    url = f'{DomainName.TYPE_API_OBJECTS.value}/{fdo_service_ref}'
    service = requests.get(url).json()

    # get the service ID
    service_id = service['serviceId']
    return service_id



import json
import uuid
from pathlib import Path
from doip_sdk import send_request

from doipy.exceptions import InvalidRequestException
from doipy.constants import DOIPOperation, ResponseStatus, DOType
from doipy.models import Authentication
from doipy.request_and_response import decode_json_response


def hello(target_id: str, ip: str, port: int):
    """
    Implements 0.DOIP/Op.Hello: An operation to allow a client to get information about the DOIP service.
    """
    # create request message
    message = {
        'targetId': f'{target_id}',
        'operationId': DOIPOperation.HELLO.value
    }
    # send request and return response
    response = send_request(ip, port, [message])
    response_decoded = decode_json_response(response)
    return response_decoded


def list_operations(target_id: str, ip: str, port: int, username: str = None, client_id: str = None,
                    password: str = None, token: str = None):
    """
    Implements 0.DOIP/Op.ListOperations: An operation to request the list of operations that can be invoked on the
    target DO.
    """
    # create request message
    message = {
        'targetId': f'{target_id}',
        'operationId': DOIPOperation.LIST_OPERATION.value
    }
    # validate authentication credentials and build message with authentication credentials, if authentication
    # credentials are provided
    if username or client_id or password or token:
        authentication = Authentication.create_instance(username, client_id, password, token)
        authentication_message = authentication.build_authentication_message()
        # concatenate messages
        message = message | authentication_message
    # send request and read response
    response = send_request(ip, port, [message])
    response_decoded = decode_json_response(response)
    return response_decoded


def create(target_id: str, ip: str, port: int, do_type: str = DOType.DO.value, do_name: str = 'Digital Object',
           bitsq: Path = None, metadata: dict = None, username: str = None, client_id: str = None, password: str = None,
           token: str = None):
    # store messages in data
    data = []

    # create first message
    message_1 = {
        'targetId': f'{target_id}',
        'operationId': DOIPOperation.CREATE.value
    }
    # validate authentication credentials and build message with authentication credentials
    authentication = Authentication.create_instance(username, client_id, password, token)
    authentication_message = authentication.build_authentication_message()

    # concatenate messages
    message_1 = message_1 | authentication_message
    data.append(message_1)

    # create second message: DO of type document in Cordra for the file which is added
    message_2 = {
        'type': do_type,
        'attributes': {
            'content': {
                'id': '',
                'name': do_name
            }
        }
    }

    # add metadata to DO
    if metadata:
        message_2['attributes']['content'] = message_2['attributes']['content'] | metadata

    if bitsq:
        # add information on files to DO
        filename = bitsq.name
        my_uuid = str(uuid.uuid4())
        message_2['elements'] = [
            {
                'id': my_uuid,
                'type': 'text/plain',
                'attributes': {
                    'filename': filename
                }
            }
        ]
        data.append(message_2)

        # third message
        message_3 = {
            'id': my_uuid
        }
        data.append(message_3)

        # send content of files
        data.append(bitsq)

    else:
        data.append(message_2)

    # send request and read response
    response = send_request(ip, port, data)
    response_decoded = decode_json_response(response)

    if response_decoded[0]['status'] == ResponseStatus.SUCCESS.value:
        return response_decoded
    raise InvalidRequestException(response)


# todo: update
"""
def update(service: str, client_id: str, password: str, do_type: str):
    # TODO fix message

    # get service settings
    target_id, ip, port = get_connection(service)

    with create_socket(ip, port) as ssl_sock:
        message = {
            'clientId': client_id,
            'targetId': target_id,
            'operationId': DOIPOperation.UPDATE.value,
            'authentication': {
                'password': password
            }
        }
        send_message(message, ssl_sock)
        string1 = f'https://cordra.testbed.pid.gwdg.de/objects/{target_id}?payload=file'
        string2 = f'https://cordra.testbed.pid.gwdg.de/objects/{target_id}'
        message = {
            'type': do_type,
            'attributes': {
                'content': {
                    'id': '',
                    'Payload': string1,
                    'Metadata': string2
                }
            }
        }
        send_message(message, ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response
"""


def retrieve(target_id: str, ip: str, port: int, file: str = None, username: str = None, client_id: str = None,
             password: str = None, token: str = None):
    """
    Implements 0.DOIP/Op.Retrieve: An operation to allow a client to get information about an (F)DO at a service.
    """
    # create message
    message = {
        'targetId': target_id,
        'operationId': DOIPOperation.RETRIEVE.value
    }
    if file:
        message['attributes'] = {
            'element': file
        }
    if username or client_id or password or token:
        # validate authentication credentials and build message with authentication credentials
        authentication = Authentication.create_instance(username, client_id, password, token)
        authentication_message = authentication.build_authentication_message()
        # concatenate messages
        message = message | authentication_message

    # send request and return response
    response = send_request(ip, port, [message])

    # decode first segment
    first_segment_json = json.loads(response.content[0])

    # if response is success and a file is returned
    if first_segment_json['status'] == ResponseStatus.SUCCESS.value and file:

        # if a filename is given, use this file name
        filename = first_segment_json.get('attributes').get('filename')
        splits = filename.split('/')
        filename = splits[len(splits)-1]
        if filename:
            f = filename
        else:
            f = 'data'
        with open(f, "wb") as binary_file:
            # Write bytes to file
            binary_file.write(response.content[1])
        return [first_segment_json]

    # if response is not success or no file is returned
    else:
        response_decoded = decode_json_response(response)
        return response_decoded


# delete
def delete(target_id: str, ip: str, port: int, username: str = None, client_id: str = None, password: str = None,
           token: str = None):
    """
    Implements 0.DOIP/Op.Delete: An operation to allow a client to delete an (F)DO at a service. This operation just
    deletes the referenced (F)DO but not any other (F)DOs which are referenced by the given (F)DO.
        """
    # create message
    message = {
        'targetId': target_id,
        'operationId': DOIPOperation.DELETE.value
    }
    # validate authentication credentials and build message with authentication credentials
    authentication = Authentication.create_instance(username, client_id, password, token)
    authentication_message = authentication.build_authentication_message()
    # concatenate messages
    message = message | authentication_message

    # send request and return response
    response = send_request(ip, port, [message])
    response_decoded = decode_json_response(response)
    return response_decoded


def search(target_id: str, ip: str, port: int, query: str, username: str = None, client_id: str = None,
           password: str = None, token: str = None):
    # create message
    message = {
        'targetId': target_id,
        'operationId': DOIPOperation.SEARCH.value,
        'attributes': {
            'query': query
        }
    }
    if username or client_id or password or token:
        # validate authentication credentials and build message with authentication credentials
        authentication = Authentication.create_instance(username, client_id, password, token)
        authentication_message = authentication.build_authentication_message()
        # concatenate messages
        message = message | authentication_message

    # send request and return response
    response = send_request(ip, port, [message])
    response_decoded = decode_json_response(response)
    return response_decoded



import json
import requests
from jsonschema import validate
from pydantic import BaseModel
from pathlib import Path

from doipy.constants import DomainName, ValidationSchemas
from doipy.exceptions import AuthenticationException


class Authentication(BaseModel):
    # optional
    username: str = None
    client_id: str = None
    password: str = None
    token: str = None

    @classmethod
    def create_instance(cls, username: str = None, client_id: str = None, password: str = None,
                        token: str = None) -> 'Authentication':
        # TODO: can the validation process be formulated as part of the DTR? -> ask Hans
        if not username and not client_id and not token:
            raise AuthenticationException('Provide token, username or client_id')

        # build the Authentication object
        authentication = cls()
        if token:
            authentication.token = token
        elif client_id:
            if password:
                authentication.client_id = client_id
                authentication.password = password
            else:
                raise AuthenticationException('Provide password')
        else:
            if password:
                authentication.username = username
                authentication.password = password
            else:
                raise AuthenticationException('Provide password')
        return authentication

    def build_authentication_message(self) -> dict:
        # create the message
        authentication_message = {}
        if self.token:
            authentication_message['authentication'] = {
                'token': self.token
            }
        elif self.client_id:
            authentication_message['clientId'] = self.client_id
            authentication_message['authentication'] = {
                'password': self.password
            }
        else:
            authentication_message['authentication'] = {
                'username': self.username,
                'password': self.password
            }
        return authentication_message


class DataAndMetadata(BaseModel):
    # optional
    data_bitsq: Path = None
    data_values: dict = None
    metadata_bitsq: Path = None
    metadata_values: dict = None

    @classmethod
    def parse(cls, user_input: dict) -> 'DataAndMetadata':
        # validate the input against the JSON input schema
        url = f'{DomainName.TYPE_API_SCHEMAS.value}/{ValidationSchemas.DATA_AND_METADATA.value}'
        input_schema = requests.get(url).json()

        validate(instance=user_input, schema=input_schema)

        do_input = cls()

        if 'data_bitsq' in user_input:
            data_bitsq = Path(user_input['data_bitsq'])
            do_input.data_bitsq = data_bitsq

        if 'metadata_bitsq' in user_input:
            metadata_bitsq = Path(user_input['metadata_bitsq'])
            do_input.metadata_bitsq = metadata_bitsq

        if 'data_values' in user_input:
            with open(user_input['data_values']) as f:
                data_values = json.load(f)
                do_input.data_values = data_values

        if 'metadata_values' in user_input:
            with open(user_input['metadata_values']) as f:
                metadata_values = json.load(f)
                do_input.metadata_values = metadata_values

        return do_input


class FdoInput(BaseModel):
    # mandatory
    fdo_service_ref: str
    fdo_profile_ref: str
    authentication: Authentication
    fdo_type_ref: str
    # optional
    fdo_rights_ref: str = None
    fdo_genre_ref: str = None
    data_and_metadata: list[DataAndMetadata] = None

    @classmethod
    def parse(cls, user_input: dict, input_schema: dict) -> 'FdoInput':

        # validate the input against the JSON input schema
        validate(instance=user_input, schema=input_schema)

        # construct the authentication object
        auth = Authentication.create_instance(user_input['FDO_Authentication'].get('username'),
                                              user_input['FDO_Authentication'].get('client_id'),
                                              user_input['FDO_Authentication'].get('password'),
                                              user_input['FDO_Authentication'].get('token'))

        # construct the fdo_input object with the mandatory references
        fdo_input = cls(fdo_service_ref=user_input['FDO_Service_Ref'], fdo_profile_ref=user_input['FDO_Profile_Ref'],
                        fdo_type_ref=user_input['FDO_Type_Ref'], authentication=auth)

        # define the optional references
        if 'FDO_Rights_Ref' in user_input:
            fdo_input.fdo_rights_ref = user_input['FDO_Rights_Ref']
        if 'FDO_Genre_Ref' in user_input:
            fdo_input.fdo_genre_ref = user_input['FDO_Genre_Ref']

        # define the data and metadata arrays which will be used to create the DOs
        if user_input['FDO_Data_and_Metadata']:
            fdo_input.data_and_metadata = []
            for item in user_input['FDO_Data_and_Metadata']:
                if item:
                    # create the structure for the data DO and for the metadata DO
                    do = DataAndMetadata.parse(item)
                    fdo_input.data_and_metadata.append(do)

        return fdo_input



from doip_sdk import send_request

from doipy.dtr_utils import get_connection
from doipy.constants import CordraOperation
from doipy.request_and_response import decode_json_response

# get service settings
target_id, ip, port = get_connection('21.T11966/service')


def get_design():
    message = {
        'targetId': f'{target_id}',
        'operationId': CordraOperation.GET_DESIGN.value
    }
    # send request and return response
    response = send_request(ip, port, [message])
    response_decoded = decode_json_response(response)
    return response_decoded


def get_init_data():
    message = {
        'targetId': f'{target_id}',
        'operationId': CordraOperation.GET_INIT_DATA.value
    }
    # send request and return response
    response = send_request(ip, port, [message])
    response_decoded = decode_json_response(response)
    return response_decoded



class AuthenticationException(Exception):
    pass


class InvalidRequestException(Exception):
    pass


class OperationNotSupportedException(Exception):
    pass


class ProfileNotSupportedException(Exception):
    pass


class HandleValueNotFoundException(Exception):
    pass


from doip_sdk import send_request

from doipy.actions.doip import create
from doipy.constants import DOType, TypeIdentifier, ResponseStatus, DOIPOperation
from doipy.dtr_utils import get_service_id, get_connection
from doipy.exceptions import InvalidRequestException
from doipy.models import FdoInput
from doipy.request_and_response import decode_json_response


def create_fdo_config_type_14(fdo_input: FdoInput):

    # get service ID, IP and port
    service_id = get_service_id(fdo_input.fdo_service_ref)
    service_id, ip, port = get_connection(service_id)

    # create first message
    message_1 = {
        'targetId': f'{service_id}',
        'operationId': DOIPOperation.CREATE.value,
    }

    # provide correct set of authentication credentials
    authentication_message = fdo_input.authentication.build_authentication_message()
    message_1 = message_1 | authentication_message

    # create second message
    message_2 = {
        # TODO: type should not be part of the DOIP message, but of the adaptor (?)
        'type': DOType.FDO.value,
        'attributes': {
            'content': {
                'id': '',
                'name': 'FAIR Digital Object',
                # FDO_Profile_Ref: mandatory
                TypeIdentifier.FDO_PROFILE_REF.value: fdo_input.fdo_profile_ref,
                # FDO_Type_Ref: mandatory
                TypeIdentifier.FDO_TYPE_REF.value: fdo_input.fdo_type_ref
            }
        }
    }
    # FDO_Rights_Ref: optional
    if fdo_input.fdo_rights_ref:
        message_2['attributes']['content'][TypeIdentifier.FDO_RIGHTS_REF.value] = fdo_input.fdo_rights_ref
    # FDO_Genre_Ref: optional
    if fdo_input.fdo_genre_ref:
        message_2['attributes']['content'][TypeIdentifier.FDO_GENRE_REF.value] = fdo_input.fdo_genre_ref

    # create the data and metadata DOs
    if fdo_input.data_and_metadata:
        data_refs = []
        metadata_refs = []
        for item in fdo_input.data_and_metadata:
            # create the data do
            if item.data_bitsq or item.data_values:
                response = create(service_id, ip, port, DOType.DO.value, 'Data-DO',
                                  item.data_bitsq, item.data_values, fdo_input.authentication.username,
                                  fdo_input.authentication.client_id, fdo_input.authentication.password,
                                  fdo_input.authentication.token)
                data_ref = response[0]['output']['id']
                data_refs.append(data_ref)
            # create the metadata do
            if item.metadata_bitsq or item.metadata_values:
                response = create(service_id, ip, port, DOType.DO.value, 'Metadata-DO',
                                  item.metadata_bitsq, item.metadata_values, fdo_input.authentication.username,
                                  fdo_input.authentication.client_id, fdo_input.authentication.password,
                                  fdo_input.authentication.token)
                metadata_ref = response[0]['output']['id']
                metadata_refs.append(metadata_ref)
        if data_refs:
            message_2['attributes']['content'][TypeIdentifier.FDO_DATA_REFS.value] = data_refs
        if metadata_refs:
            message_2['attributes']['content'][TypeIdentifier.FDO_MD_REFS.value] = metadata_refs

    # send request and read response
    data = [message_1, message_2]
    response = send_request(ip, port, data)
    response_decoded = decode_json_response(response)

    if response_decoded[0]['status'] == ResponseStatus.SUCCESS.value:
        return response_decoded
    raise InvalidRequestException(response)
